%%%-------------------------------------------------------------------
%%% File    : hibernate.erl
%%% Author  :  <>
%%% Description : 
%%%
%%% Created : 30 Jun 2009 by  <>
%%%-------------------------------------------------------------------
-module(hibernate).

-behaviour(gen_server_light).

%% API
-export([start/1, stop/1, test/0, call_and_hibernate/4]).

%% gen_server callbacks + handle_reply/3
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
	 terminate/2, code_change/3, handle_reply/2]).

test() ->
	S1 = hibernating_server,
	S2 = remote_server,
	?MODULE:start(S1),
	?MODULE:start(S2),
	io:format("test 1: normal request...~n"),
	call_and_hibernate(S1, S2, echo_me, infinity),
	io:format("~ntest 2: timeout failure...~n"),
	call_and_hibernate(S1, S2, {stall,2000}, 500),
	io:format("~ntest 3: showing how \"lost\" replies are handled...~n"),
	call_and_hibernate(S1, S2, priority_over_timed_out_reply, infinity),
	io:format("~ntest 4: comparing the total_heap_size between client and server...~n"),
	call_and_hibernate(S1, S2, used_memory, infinity),
	?MODULE:stop(S1),
	?MODULE:stop(S2),
	test_completed.
	

%%====================================================================
%% API
%%====================================================================
start(Name) ->
    %%gen_server_light:start({local, Name}, ?MODULE, [], [{debug, [trace]}]).
	gen_server_light:start({local, Name}, ?MODULE, [], []).

stop(Server) ->
    gen_server_light:call(Server, stop).

call_and_hibernate(S1, S2, Req, ReqTimeout) ->
	gen_server_light:call(S1, {call_and_hibernate, S2, Req, ReqTimeout}).

%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%--------------------------------------------------------------------
init([]) ->
    {ok, []}.

%%--------------------------------------------------------------------
%% Function: %% handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call({call_and_hibernate, Server, Req, ReqTimeout}, From, _State) ->
	%% We'll store the caller id to use later on when we get a reply
	NewState = From,
	io:format("S1: routing the request to S2!~n"),
    {noreply, NewState, {call_and_hibernate, {Server, Req, ReqTimeout}}};
	
handle_call({stall,ReplyDelay}, _From, State) ->
	io:format("S2: delaying my reply for ~p milliseconds!~n", [ReplyDelay]),
    timer:sleep(ReplyDelay),
	io:format("S2: replying now (hope S1 is still waiting for me)!~n"),
    {reply, stalled_reply, State};

handle_call(used_memory, From, State) ->
	{S1Pid, _CallRef} = From,
	S1Mem = erlang:process_info(S1Pid, total_heap_size),
    S2Mem = erlang:process_info(self(), total_heap_size),
	Reply = {S1Mem, S2Mem},
    {reply, Reply, State};
	
handle_call(Request, _From, State) ->
    {reply, {echo, Request}, State}.

%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
	{stop, normal, State};
handle_cast(_Msg, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------
handle_info({Ref, UnexpectedReply}, State) when is_reference(Ref) ->
	io:format("S1: got an unexpected reply (handled by handle_info/2) = ~p!~n", [UnexpectedReply]),
    {noreply, State};
handle_info(_Info, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_reply(Msg, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Msg :: {ok, Reply} | {nodedown, Node} | {nodedown, Node, Reason} | timeout
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------
handle_reply({ok, Reply}, State) ->
    io:format("S1: reply from S2 = ~p~n", [Reply]),
	Caller = State,
	gen_server_light:reply(Caller, Reply),
	{noreply, []};
handle_reply({nodedown, _Node}, State) ->
    io:format("S1: S2's node is down!~n"),
	Caller = State,
	gen_server_light:reply(Caller, node_down),
	{noreply, []};
handle_reply({nodedown, _Node, Reason}, State) ->
    io:format("S1: S2's node is down due to ~p.~n", [Reason]),
	Caller = State,
	gen_server_light:reply(Caller, {node_down, Reason}),
	{noreply, []};
handle_reply(timeout, State) ->
    io:format("S1: timeout from S2 before getting a reply (keep an eye for late replies that might fill up the message queue).~n"),
	Caller = State,
	gen_server_light:reply(Caller, timeout),
	{noreply, []};
handle_reply(Msg, State) ->
    io:format("S1: unexpected reply from S2 = ~p~n", [Msg]),
	Caller = State,
	gen_server_light:reply(Caller, unexpected_reply),
	{noreply, []}.

%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(_Reason, _State) ->
    ok.

%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%% FAILED

%% H1 SEND
%% *DBG* h1 got call {call_and_hibernate,h2,used_memory,infinity} from <0.517.0>
%% S1: routing the request to S2!
%% *DBG* h1 new state {<0.517.0>,#Ref<0.0.0.6915>}

%% DELAYED REPLY
%% S2: replying now (hope S1 is still waiting for me)!
%% *DBG* h2 sent stalled_reply to <0.504.0>, new state []
%% *DBG* h1 got {#Ref<0.0.0.6907>,stalled_reply}

%% S2: NEW REQUEST ARRIVAL
%% \*DBG* h2 got call used_memory from <0.504.0>

%% S1: STALLED_REPLY RECEIVED ON handle_info/2
%% S1: only know did I got the delayed message from S2 = {#Ref<0.0.0.6907>,
%%                                                        stalled_reply}!

%% S2: REPLYING TO THE NEW REQUEST
%% *DBG* h2 sent {{total_heap_size,610},{total_heap_size,1597}} to <0.504.0>, new state []

%% S1: EXITING handle_info/2 (with the original state)
%% \*DBG* h1 new state {<0.517.0>,#Ref<0.0.0.6915>}

%% FOUND THE PROBLEM!! The REPLY ISN'T BEING WRAPPED IN THE '$gen_reply' tuple
%% so it ends up being delivered to the handle_info/2 function instead of the handle_reply/2 one.

%% \*DBG* h1 got {#Ref<0.0.0.6919>,{{total_heap_size,610},{total_heap_size,1597}}}
%% S1: only know did I got the delayed message from S2 = {#Ref<0.0.0.6919>,
%%                                                       {{total_heap_size,610},
%%                                                        {total_heap_size,
%%                                                         1597}}}!

%%*DBG* h1 new state {<0.517.0>,#Ref<0.0.0.6915>}
%%** exception exit: {timeout,
%%                       {gen_server_light,call,
%%                           [h1,{call_and_hibernate,h2,used_memory,infinity}]}}
%%     in function  gen_server_light:call/2
%%     in call from hibernate:test/0



%% CORRECT

%% H1 SEND
%% \*DBG* h1 got call {call_and_hibernate,h2,used_memory,infinity} from <0.517.0>
%% S1: routing the request to S2!
%% \*DBG* h1 new state {<0.517.0>,#Ref<0.0.0.6825>}

%% \*DBG* h2 got call used_memory from <0.504.0>
%% \*DBG* h2 sent {{total_heap_size,217},{total_heap_size,610}} to <0.504.0>, new state []
%% \*DBG* h1 got {'$gen_reply',{ok,{{total_heap_size,217},{total_heap_size,610}}}}
%% S1: reply from S2 = {{total_heap_size,217},{total_heap_size,610}}
%% \*DBG* h1 new state []
%% {{total_heap_size,217},{total_heap_size,610}}

%% Interweaving/deleting equals:

%% \*DBG* h1 new state {<0.517.0>,#Ref<0.0.0.6825>}
%% \*DBG* h2 got call used_memory from <0.504.0>
%% \*DBG* h2 sent {{total_heap_size,217},{total_heap_size,610}} to <0.504.0>, new state []
%% \*DBG* h1 got {'$gen_reply',{ok,{{total_heap_size,217},{total_heap_size,610}}}}
%% S1: reply from S2 = {{total_heap_size,217},{total_heap_size,610}}
%% \*DBG* h1 new state []
%% {{total_heap_size,217},{total_heap_size,610}}